using cuqmbr.TravelGuide.Application.Common.Exceptions;
using Microsoft.AspNetCore.Mvc;
using Microsoft.Extensions.Localization;
using System.Reflection;

namespace cuqmbr.TravelGuide.HttpApi.Middlewares;

public class GlobalExceptionHandlerMiddleware : IMiddleware
{
    private readonly Dictionary<Type, Func<HttpContext, Exception, Task>> _exceptionHandlers;
    private readonly ILogger _logger;
    private readonly IStringLocalizer _localizer;

    public GlobalExceptionHandlerMiddleware(
        ILogger<GlobalExceptionHandlerMiddleware> logger,
        IStringLocalizer localizer)
    {
        _exceptionHandlers = new()
        {
            // Request object validation
            { typeof(ValidationException), HandleValidationException },
            // Authentication
            // TODO: Remove UnAuthorizedException, isnt used
            { typeof(UnAuthorizedException), HandleUnAuthorizedException },
            { typeof(AuthenticationException), HandleAuthenticationException },
            { typeof(RegistrationException), HandleRegistrationException },
            { typeof(LoginException), HandleLoginException },
            // Authorization
            { typeof(ForbiddenException), HandleForbiddenException },
            // Data access
            { typeof(DuplicateEntityException), HandleDuplicateEntityException },
            { typeof(NotFoundException), HandleNotFoundException },
        };

        _logger = logger;
        _localizer = localizer;
    }

    public async Task InvokeAsync(HttpContext context, RequestDelegate next)
    {
        try
        {
            await next(context);
        }
        catch (Exception exception)
        {
            var exceptionType = exception.GetType();

            if (_exceptionHandlers.ContainsKey(exceptionType))
            {
                _logger.LogInformation(
                        "Interrupted with planned exception {ExceptionType}.",
                        exception.GetType());

                context.Response.ContentType = "application/problem+json";

                await _exceptionHandlers[exceptionType].Invoke(context, exception);
                return;
            }

            _logger.LogError(
                exception,
                "Interrupted with unhandled exception {ExceptionType}.",
                exception.GetType());

            await HandleUnhandledException(context, exception);
        }
    }

    private async Task HandleValidationException(
        HttpContext context,
        Exception exception)
    {
        var ex = (ValidationException)exception;

        context.Response.StatusCode = StatusCodes.Status400BadRequest;

        await context.Response.WriteAsJsonAsync(new HttpValidationProblemDetails(ex.Errors)
        {
            Status = StatusCodes.Status400BadRequest,
            Type = "https://datatracker.ietf.org/doc/html/rfc7231#section-6.5.1",
            Title = _localizer["ExceptionHandling.ValidationException.Title"],
            Detail = _localizer["ExceptionHandling.ValidationException.Detail"]
        });
    }

    private async Task HandleUnAuthorizedException(
        HttpContext context,
        Exception exception)
    {
        context.Response.StatusCode = StatusCodes.Status401Unauthorized;

        await context.Response.WriteAsJsonAsync(new ProblemDetails()
        {
            Status = StatusCodes.Status401Unauthorized,
            Type = "https://datatracker.ietf.org/doc/html/rfc7235#section-3.1",
            Title = _localizer["ExceptionHandling.UnAuthorizedException.Title"],
            Detail = _localizer["ExceptionHandling.UnAuthorizedException.Detail"]
        });
    }

    private async Task HandleAuthenticationException(
        HttpContext context,
        Exception exception)
    {
        context.Response.StatusCode = StatusCodes.Status400BadRequest;

        await context.Response.WriteAsJsonAsync(new ProblemDetails()
        {
            Status = StatusCodes.Status400BadRequest,
            Type = "https://datatracker.ietf.org/doc/html/rfc7231#section-6.5.1",
            Title = _localizer["ExceptionHandling.AuthenticationException.Title"],
            Detail = _localizer["ExceptionHandling.AuthenticationException.Detail"]
        });
    }

    private async Task HandleRegistrationException(
        HttpContext context,
        Exception exception)
    {
        context.Response.StatusCode = StatusCodes.Status400BadRequest;

        await context.Response.WriteAsJsonAsync(new ProblemDetails()
        {
            Status = StatusCodes.Status400BadRequest,
            Type = "https://datatracker.ietf.org/doc/html/rfc7231#section-6.5.1",
            Title = _localizer["ExceptionHandling.RegistrationException.Title"],
            Detail = _localizer["ExceptionHandling.RegistrationException.Detail"]
        });
    }

    private async Task HandleLoginException(
        HttpContext context,
        Exception exception)
    {
        context.Response.StatusCode = StatusCodes.Status400BadRequest;

        await context.Response.WriteAsJsonAsync(new ProblemDetails()
        {
            Status = StatusCodes.Status400BadRequest,
            Type = "https://datatracker.ietf.org/doc/html/rfc7231#section-6.5.1",
            Title = _localizer["ExceptionHandling.LoginException.Title"],
            Detail = _localizer["ExceptionHandling.LoginException.Detail"]
        });
    }

    private async Task HandleForbiddenException(
        HttpContext context,
        Exception exception)
    {
        context.Response.StatusCode = StatusCodes.Status403Forbidden;

        await context.Response.WriteAsJsonAsync(new ProblemDetails()
        {
            Status = StatusCodes.Status403Forbidden,
            Type = "https://datatracker.ietf.org/doc/html/rfc7231#section-6.5.3",
            Title = _localizer["ExceptionHandling.ForbiddenException.Title"],
            Detail = _localizer["ExceptionHandling.ForbiddenException.Detail"]
        });
    }

    private async Task HandleDuplicateEntityException(
        HttpContext context,
        Exception exception)
    {
        context.Response.StatusCode = StatusCodes.Status400BadRequest;

        await context.Response.WriteAsJsonAsync(new ProblemDetails()
        {
            Status = StatusCodes.Status400BadRequest,
            Type = "https://datatracker.ietf.org/doc/html/rfc7231#section-6.5.1",
            Title = _localizer["ExceptionHandling.DuplicateEntityException.Title"],
            Detail = _localizer["ExceptionHandling.DuplicateEntityException.Detail"]
        });
    }

    private async Task HandleNotFoundException(
        HttpContext context,
        Exception exception)
    {
        context.Response.StatusCode = StatusCodes.Status404NotFound;

        await context.Response.WriteAsJsonAsync(new ProblemDetails()
        {
            Status = StatusCodes.Status404NotFound,
            Type = "https://datatracker.ietf.org/doc/html/rfc7231#section-6.5.4",
            Title = _localizer["ExceptionHandling.NotFoundException.Title"],
            Detail = _localizer["ExceptionHandling.NotFoundException.Detail"]
        });
    }

    private async Task HandleUnhandledException(HttpContext context, Exception exception)
    {
        context.Response.StatusCode = StatusCodes.Status500InternalServerError;

        await context.Response.WriteAsJsonAsync(new ProblemDetails()
        {
            Status = StatusCodes.Status500InternalServerError,
            Type = "https://datatracker.ietf.org/doc/html/rfc7231#section-6.6.1",
            Title = _localizer["ExceptionHandling.UnhandledException.Title"],
            Detail = _localizer["ExceptionHandling.UnhandledException.Detail"]
        });
    }

    // class ProblemDetailsWithTraceId : ProblemDetails
    // {
    //     public string TraceId { get; init; } = Activity.Current?.TraceId.ToString();
    // }
}

