using System.Security.Cryptography;
using cuqmbr.TravelGuide.Application.Authenticaion;
using cuqmbr.TravelGuide.Application.Common.Interfaces.Services;
using Microsoft.AspNetCore.Identity;
using Microsoft.Extensions.Options;
using IdentityUser = cuqmbr.TravelGuide.Identity.Models.IdentityUser;
using IdentityRole = cuqmbr.TravelGuide.Identity.Models.IdentityRole;
using cuqmbr.TravelGuide.Identity.Models;
using System.Security.Claims;
using System.Text;
using Microsoft.IdentityModel.Tokens;
using cuqmbr.TravelGuide.Application.Common.Models;
using System.IdentityModel.Tokens.Jwt;
using cuqmbr.TravelGuide.Application.Common.Exceptions;

namespace cuqmbr.TravelGuide.Identity.Services;

public class JwtAuthenticationService : AuthenticationService
{
    private readonly ConfigurationOptions _configuration;

    private readonly UserManager<IdentityUser> _userManager;
    private readonly RoleManager<IdentityRole> _roleManager;

    public JwtAuthenticationService(
        UserManager<IdentityUser> userManager,
        RoleManager<IdentityRole> roleManager,
        IOptions<ConfigurationOptions> configuration)
    {
        _userManager = userManager;
        _userManager.UserValidators.Clear();
        _userManager.PasswordValidators.Clear();

        _roleManager = roleManager;

        _configuration = configuration.Value;
    }

    public async Task RegisterAsync(
        string email, 
        string password,
        CancellationToken cancellationToken)
    {
        var userWithSameEmail = await _userManager.FindByEmailAsync(email);
        if (userWithSameEmail is not null)
        {
            throw new RegistrationException("User with given email already registered.");
        }

        var newUser = new IdentityUser
        {
            // Id = Guid.NewGuid().ToString(),
            Email = email,
            RefreshTokens = default!
        };

        var createUserResult = await _userManager.CreateAsync(newUser, password);

        if (createUserResult.Errors.Any())
        {
            var errorMessage = String.Join("\n",
                    createUserResult.Errors.Select(e => e.Description));
            throw new AuthenticationException(errorMessage);
        }

        var addToRoleResult =
            await _userManager.AddToRoleAsync(newUser, Application.Common.Models.IdentityRole.User.Name);

        if (addToRoleResult.Errors.Any())
        {
            var errorMessage = String.Join("\n",
                    addToRoleResult.Errors.Select(e => e.Description));
            throw new AuthenticationException(errorMessage);
        }
    }

    public async Task<TokensModel> LoginAsync(
        string email,
        string password,
        CancellationToken cancellationToken)
    {
        var user = await _userManager.FindByEmailAsync(email);
        if (user is null)
        {
            throw new LoginException("No users registered with given email.");
        }

        var isPasswordCorrect = await _userManager.CheckPasswordAsync(user, password);
        if (!isPasswordCorrect)
        {
            throw new LoginException("Given password is incorrect.");
        }

        var jwtSecurityToken = await CreateJwtAsync(user, cancellationToken);
        var accessToken = new JwtSecurityTokenHandler().WriteToken(jwtSecurityToken);

        var refreshToken = user.RefreshTokens.FirstOrDefault(t => t.IsActive);
        if (refreshToken is null)
        {
            refreshToken = CreateRefreshToken();
            refreshToken.IdentityUserId = user.Id;
            user.RefreshTokens.Add(refreshToken);
            var result = await _userManager.UpdateAsync(user);
        }

        return new TokensModel(accessToken, refreshToken.Value);
    }

    public async Task<TokensModel> RenewAccessTokenAsync(
        string refreshToken,
        CancellationToken cancellationToken)
    {
        var user = _userManager.Users.SingleOrDefault(u =>
                u.RefreshTokens.Any(rt => rt.Value == refreshToken));
        if (user is null)
        {
            throw new AuthenticationException($"Refresh token was not found.");
        }

        var refreshTokenObject = user.RefreshTokens.Single(rt =>
                rt.Value == refreshToken);
        if (!refreshTokenObject.IsActive)
        {
            throw new AuthenticationException("Refresh token is inactive.");
        }

        var jwtSecurityToken = await CreateJwtAsync(user, cancellationToken);
        var accessToken = new JwtSecurityTokenHandler().WriteToken(jwtSecurityToken);

        return new TokensModel(accessToken, refreshToken);
    }

    public async Task RevokeRefreshTokenAsync(
        string refreshToken,
        CancellationToken cancellationToken)
    {
        var user = _userManager.Users.SingleOrDefault(u =>
                u.RefreshTokens.Any(t => t.Value == refreshToken));
        if (user is null)
        {
            throw new AuthenticationException("Invalid refreshToken");
        }

        var refreshTokenObject = user.RefreshTokens.Single(x =>
                x.Value == refreshToken);
        if (!refreshTokenObject.IsActive)
        {
            throw new AuthenticationException("RefreshToken already revoked");
        }

        refreshTokenObject.RevokationTimestamp = DateTimeOffset.UtcNow;
        await _userManager.UpdateAsync(user);
    }

    private async Task<JwtSecurityToken> CreateJwtAsync(
        IdentityUser user,
        CancellationToken cancellationToken)
    {
        var userClaims = await _userManager.GetClaimsAsync(user);

        var roles = await _userManager.GetRolesAsync(user);
        var roleClaims = new List<Claim>();
        foreach (var role in roles)
        {
            roleClaims.Add(new Claim("roles", role));
        }

        var claims = new List<Claim>()
        {
            new Claim(JwtRegisteredClaimNames.Sub, user.Id.ToString()),
            new Claim(JwtRegisteredClaimNames.Email, user.Email)
        }
        .Union(userClaims)
        .Union(roleClaims);

        var expirationDateTimeUtc = DateTime.UtcNow.Add(
            _configuration.JsonWebToken.AccessTokenValidity);

        var symmetricSecurityKey = new SymmetricSecurityKey(
            Encoding.UTF8.GetBytes(_configuration.JsonWebToken.IssuerSigningKey));
        var signingCredentials = new SigningCredentials(
            symmetricSecurityKey, SecurityAlgorithms.HmacSha256);

        var jwtSecurityToken = new JwtSecurityToken(
            issuer: _configuration.JsonWebToken.Issuer,
            audience: _configuration.JsonWebToken.Audience,
            claims: claims,
            expires: expirationDateTimeUtc,
            signingCredentials: signingCredentials);

        return jwtSecurityToken;
    }

    private RefreshToken CreateRefreshToken()
    {
        var randomNumber = new byte[32];

        using var rng = RandomNumberGenerator.Create();
        rng.GetNonZeroBytes(randomNumber);

        return new RefreshToken
        {
            // Id = Guid.NewGuid().ToString(),
            Value = Convert.ToBase64String(randomNumber),
            CreationTimestamp = DateTimeOffset.UtcNow,
            ExpirationTimestamp = DateTimeOffset.UtcNow.Add(
                _configuration.JsonWebToken.RefreshTokenValidity)
        };
    }
}
