using System.Linq.Expressions;
using cuqmbr.TravelGuide.Application.Common.Interfaces.Persistence.Repositories;
using cuqmbr.TravelGuide.Application.Common.Models;
using cuqmbr.TravelGuide.Domain.Entities;
using Microsoft.EntityFrameworkCore;

namespace cuqmbr.TravelGuide.Persistence.PostgreSql.Repositories;

public abstract class PostgreSqlBaseRepository<TEntity> : BaseRepository<TEntity>
    where TEntity : EntityBase
{
    protected readonly DbSet<TEntity> _dbSet;

    public PostgreSqlBaseRepository(PostgreSqlDbContext dbContext)
    {
        _dbSet = dbContext.Set<TEntity>();
    }

    public async Task<TEntity> AddOneAsync(
        TEntity entity,
        CancellationToken cancellationToken)
    {
        await _dbSet.AddAsync(entity, cancellationToken);
        return entity;
    }

    public async Task<TEntity?> GetOneAsync(
        Expression<Func<TEntity, bool>> predicate,
        CancellationToken cancellationToken)
    {
        return await _dbSet.SingleOrDefaultAsync(predicate);
    }

    public async Task<TEntity?> GetOneAsync(
        Expression<Func<TEntity, bool>> predicate,
        Expression<Func<TEntity, object>> includeSelector,
        CancellationToken cancellationToken)
    {
        return 
            await _dbSet
                .Include(includeSelector)
                .SingleOrDefaultAsync(predicate);
    }

    public async Task<PaginatedList<TEntity>> GetPageAsync(
        int pageNumber, int pageSize,
        CancellationToken cancellationToken)
    {
        var count = await _dbSet.CountAsync();

        var entities = 
            await _dbSet
                .Skip((pageNumber - 1) * pageSize).Take(pageSize)
                .ToListAsync();

        return new PaginatedList<TEntity>(
            entities, count,
            pageNumber, pageSize);
    }

    public async Task<PaginatedList<TEntity>> GetPageAsync(
        Expression<Func<TEntity, object>> includeSelector,
        int pageNumber, int pageSize,
        CancellationToken cancellationToken)
    {
        var count = await _dbSet.CountAsync();

        var entities =
            await _dbSet
                .Skip((pageNumber - 1) * pageSize).Take(pageSize)
                .Include(includeSelector)
                .ToListAsync();

        return new PaginatedList<TEntity>(
            entities, count,
            pageNumber, pageSize);
    }

    public async Task<PaginatedList<TEntity>> GetPageAsync(
        Expression<Func<TEntity, bool>> predicate,
        int pageNumber, int pageSize,
        CancellationToken cancellationToken)
    {
        var count = await _dbSet.Where(predicate).CountAsync();

        var entities =
            await _dbSet
                .Where(predicate)
                .Skip((pageNumber - 1) * pageSize).Take(pageSize)
                .ToListAsync();

        return new PaginatedList<TEntity>(
            entities, count,
            pageNumber, pageSize);
    }

    public async Task<PaginatedList<TEntity>> GetPageAsync(
        Expression<Func<TEntity, bool>> predicate,
        Expression<Func<TEntity, object>> includeSelector,
        int pageNumber, int pageSize,
        CancellationToken cancellationToken)
    {
        var count = await _dbSet.Where(predicate).CountAsync();

        var entities =
            await _dbSet
                .Where(predicate)
                .Skip((pageNumber - 1) * pageSize).Take(pageSize)
                .Include(includeSelector)
                .ToListAsync();

        return new PaginatedList<TEntity>(
            entities, count,
            pageNumber, pageSize);
    }
    
    public Task<TEntity> UpdateOneAsync(
        TEntity entity,
        CancellationToken cancellationToken)
    {
        _dbSet.Update(entity);
        return Task.FromResult(entity);
    }

    public Task DeleteOneAsync(
        TEntity entity,
        CancellationToken cancellationToken)
    {
        _dbSet.Remove(entity);
        return Task.CompletedTask;
    }

}
